%option noyywrap
%option nounput 
%option outfile="diameter_load_dictionary.c"
%option prefix="Diameter_Load_"
%option never-interactive
%option caseless

%{
/* diameter_load_dictionary.lex
* an XML dissector for ethereal 
* DTD Preparser - import a dtd file into a GString
*                 #including files, removing comments
*                 and resolving %entities;
* 
* Copyright 2004, Luis E. Garcia Ontanon <luis.ontanon@gmail.com>
*
* $Id: xml_lexer.l 13639 2005-03-06 21:57:25Z lego $
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <glib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#define MAX_INCLUDE_DEPTH 10
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int include_stack_ptr = 0;

#define ECHO g_string_append(current,yytext);

GString* current;
gchar* file;
gchar* incfile;
GString* output;
GHashTable* entities;
gchar* entity_name;
GString* error;

static gchar* replace_entity(gchar* s);

		/* [:blank:]+file[:blank:]*=[:blank:]*["] */

%}

xmpli_start "<?"
xmpli_stop  "?>"
comment_start "<!--"
comment_stop "-->"
special_start "<!"
special_stop ">"

entity_start     "<!"[[:blank:]\n]*entity[[:blank:]\n]*"%"
system_start	system[[:blank:]\n]*["]
filename		[^"]+

name [A-Za-z][-:A-Za-z0-9_]*

quote "\""
percent [%]
escaped_quote "\\\""
non_quote [^"%]+

entity        %[A-Za-z][-A-Za-z0-9_]*;

whitespace [[blank:]\n]+

%START OUTSIDE IN_COMMENT IN_ENTITY NAMED_ENTITY IN_QUOTE ENTITY_DONE INCLUDE_TAG FILENAME INCLUDED
%%

{entity}						if (current) g_string_append(current,replace_entity(yytext));
{whitespace}					if (current) g_string_append(current," ");

<OUTSIDE>{include_tag_start}	{ current = NULL; g_message("tag_start"); BEGIN INCLUDE_TAG; }
<INCLUDE_TAG>{include_filename}	{ incfile = g_strdup(yytext); g_message("filename: %s",incfile); BEGIN FILENAME; }

<FILENAME>{include_stop}	{ 

	current = output;

	if ( include_stack_ptr >= MAX_INCLUDE_DEPTH )
		g_error("dtd_preparse: include files nested to deeply");

	include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;

	yyin = fopen( incfile, "r" );

	if (!yyin) {
		yy_delete_buffer( YY_CURRENT_BUFFER );
		
		yy_switch_to_buffer(
			 include_stack[--include_stack_ptr] );

		if (errno)
			g_string_sprintfa(error, "preparse_dtd: Could not open file: '%s', error: %s", incfile, strerror(errno) );
			
		g_free(incfile);
		incfile = NULL;
		
		
		BEGIN INCLUDE_TAG;
	} else {
		yy_switch_to_buffer(yy_create_buffer( yyin, YY_BUF_SIZE ) );
		BEGIN OUTSIDE;
	}
}

<<EOF>> {
	g_warning("EOF: %s",incfile);
	
	if ( --include_stack_ptr < 0 ) {
		yyterminate();
	} else {
		yy_delete_buffer( YY_CURRENT_BUFFER );
		
		yy_switch_to_buffer(
			 include_stack[include_stack_ptr] );
	}
	
	BEGIN INCLUDED;
}

<INCLUDED>{include_stop}		{
	if (incfile) {
		g_free(incfile);
		incfile = NULL;
	} else { 
		g_string_sprintfa(error,"dtd_preparse: at xmpli_stop in INCLUDE_TAG: incfile does not exists\n");
	}
	
	current = output;

	BEGIN OUTSIDE;
}

<OUTSIDE>{comment_start}		{ current = NULL; BEGIN IN_COMMENT; }
<IN_COMMENT>[^-]?				|
<IN_COMMENT>[-]					;
<IN_COMMENT>{comment_stop}		{ current = output; BEGIN OUTSIDE; }

<OUTSIDE>{entity_start}			{ BEGIN IN_ENTITY; }
<IN_ENTITY>{name}				{ entity_name = g_strdup_printf("%%%s;",yytext); BEGIN NAMED_ENTITY; }

<NAMED_ENTITY>{quote}			{ current = g_string_new(""); BEGIN IN_QUOTE; }
<NAMED_ENTITY>{system}			{ current = g_string_new(""); BEGIN IN_SYSTEM; }

<IN_SYSTEM>{system}			{ current = g_string_new(""); BEGIN ; }

<IN_QUOTE>{quote}				{ g_hash_table_insert(entities,entity_name,current);  BEGIN ENTITY_DONE; }
<IN_QUOTE>{quote}				{ g_hash_table_insert(entities,entity_name,current);  BEGIN ENTITY_DONE; }
<IN_QUOTE>{percent}				|
<IN_QUOTE>{non_quote}			|
<IN_QUOTE>{escaped_quote}		g_string_append(current,yytext);
<ENTITY_DONE>{special_stop}		{ current = output; BEGIN OUTSIDE; }

%%

static gchar* replace_entity(gchar* entity) {
	GString* replacement = g_hash_table_lookup(entities,entity);

	if (replacement) {
		return replacement->str;
	} else {
		g_string_sprintfa(error,"dtd_preparse: in file '%s': %s does not exists\n", file, entity);
		return "";
	}
	
}

static gboolean free_gstring_hash_items(gpointer k,gpointer v,gpointer p _U_) {
	g_free(k);
	g_string_free(v,TRUE);
	return TRUE;
}

extern GString* preparse_dtd(gchar* filename, GString* err) {

	yyin = fopen(filename,"r");
	
	if (!yyin) {
		if (err)
			g_string_sprintfa(err, "preparse_dtd: Could not open file: '%s', error: %s",filename,strerror(errno));
			
		return NULL;
	}

	file = filename;
	error = err;
	entities = g_hash_table_new(g_str_hash,g_str_equal);
	current = output = g_string_new("");
		
	BEGIN OUTSIDE;
	
	yylex();
		
	yyrestart(NULL);

	g_hash_table_foreach_remove(entities,free_gstring_hash_items,NULL);
	g_hash_table_destroy(entities);
		
	return output;
}

#ifdef STANDALONE_DTD_PREPARSER
int main(int argc, const char** argv) {
#else
int test_dtd_preparse(int argc, char** argv) {
#endif

	GString* err = g_string_new("");
	GString* preparsed;
	int i;
	FILE* out = stdout;
	
	switch (argc) {
	case 3:
		out = fopen(argv[2],"w");
		
		if ( out == NULL ) {
			g_error("Could not open file: '%s', error: %s",argv[2],strerror(errno));
			return 3;
		}
	case 2:
		preparsed = preparse_dtd(argv[1],error);
		
		if ( preparsed )
			fprintf(out,"%s",preparsed->str);

		fclose(out);

		if ( err->len ) {
			g_error("%s",err->str);
			return 1;
		}
		
		return 0;
	default:
		g_message("usage: %s config.dtd [preparsed.dtd]",argv[0]);
		return 2;
	}
	
}

